# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import DashboardInfoClass
from .....schema_classes import DashboardUsageStatisticsClass
from .....schema_classes import DashboardUserUsageCountsClass
from .....schema_classes import EditableDashboardPropertiesClass


DashboardInfo = DashboardInfoClass
DashboardUsageStatistics = DashboardUsageStatisticsClass
DashboardUserUsageCounts = DashboardUserUsageCountsClass
EditableDashboardProperties = EditableDashboardPropertiesClass

# fmt: on
